using System;
using System.IO;
using System.Collections.Generic;

namespace NCMemoryStream
{
    class NonContiguousMemoryStream : Stream
    {
        readonly int bufferStoreSize = 64 * 1024;
        List<byte[]> buffersStore = new List<byte[]>();
        long position;
        long length;
        bool disposed;

        public NonContiguousMemoryStream()
        {
        }

        public NonContiguousMemoryStream(byte[] buffer)
        {
            if (buffer == null)
            {
                throw new ArgumentNullException(typeof(NonContiguousMemoryStream).ToString());
            }

            Write(buffer, 0, buffer.Length);
        }

        public override void WriteByte(byte value)
        {
            Write(new byte[] { value }, 0 , 1);
        }

        public void WriteTo(Stream stream)
        {
            if (disposed)
            {
                throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
            }

            if (stream == null)
            {
                throw new ArgumentNullException("stream");
            }

            int countMinusOne = buffersStore.Count - 1;

            if (countMinusOne >= 0)
            {
                for (int i = 0; i < countMinusOne; i++)
                {
                    stream.Write(buffersStore[i], 0, bufferStoreSize);
                }

                int remaining = (int)(length % bufferStoreSize);

                if (remaining == 0)
                {
                    remaining = bufferStoreSize;
                }

                stream.Write(buffersStore[countMinusOne], 0, remaining);
            }
        }

        public override bool CanRead
        {
            get
            {
                if (disposed)
                {
                    throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
                }

                return true;
            }
        }

        public override bool CanSeek
        {
            get
            {
                if (disposed)
                {
                    throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
                }

                return true;
            }
        }

        public override bool CanWrite
        {
            get
            {
                if (disposed)
                {
                    throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
                }

                return true;
            }
        }

        public override void Flush()
        {
        }

        public override long Length
        {
            get
            {
                if (disposed)
                {
                    throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
                }

                return length;
            }
        }

        public override long Position
        {
            get
            {
                if (disposed)
                {
                    throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
                }

                return position;
            }

            set
            {
                if (disposed)
                {
                    throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
                }

                if (value < 0 || (length > 0 && value >= length))
                {
                    throw new InvalidDataException("Position out of range");
                }

                position = value;
            }
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            if (disposed)
            {
                throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
            }

            if (buffer == null)
            {
                throw new ArgumentNullException("buffer");
            }

            if (offset < 0 || offset >= buffer.Length)
            {
                throw new ArgumentOutOfRangeException("offset", offset, "0 <= offset < buffer.Length");
            }

            if (count < 0 || count > buffer.Length)
            {
                throw new ArgumentOutOfRangeException("count", count, "0 <= count < buffer.Length");
            }

            if ((offset + count) > buffer.Length)
            {
                throw new ArgumentOutOfRangeException("offset + count", (offset + count), "(offset + count) <= buffer.Length");
            }

            int end = offset + count;
            long remaining = length - position;

            if (remaining < end)
            {
                end = (int)remaining;
            }

            int startingOffset = offset;
            int buffersStoreIndex = (int)(position / bufferStoreSize);
            int bufferStoreOffset = (int)(position % bufferStoreSize);
            byte[] bufferStore = buffersStore[buffersStoreIndex];

            while ((offset < end) && (position < length))
            {
                int run = Math.Min(end - offset, bufferStoreSize - bufferStoreOffset);
                Buffer.BlockCopy(bufferStore, bufferStoreOffset, buffer, offset, run);
                offset += run;
                bufferStoreOffset += run;
                position += run;

                if (bufferStoreOffset == bufferStoreSize && position < length)
                {
                    buffersStoreIndex += 1;
                    bufferStore = buffersStore[buffersStoreIndex];
                    bufferStoreOffset = 0;
                }
            }

            return (offset - startingOffset);
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            if (disposed)
            {
                throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
            }

            if (offset > int.MaxValue)
            {
                throw new ArgumentOutOfRangeException("offset", "offset > Int32.MaxValue");
            }

            long newPosition = position;

            switch (origin)
            {
                case SeekOrigin.Begin:
                    if (offset < 0)
                    {
                        throw new IOException("Seek before Begin");
                    }

                    newPosition = offset;
                    break;

                case SeekOrigin.Current:
                    if ((position + offset) < 0)
                    {
                        throw new IOException("Seek before Begin");
                    }

                    newPosition += offset;
                    break;

                case SeekOrigin.End:
                    if ((length + offset) < 0)
                    {
                        throw new IOException("Seek before Begin");
                    }

                    newPosition = length + offset;
                    break;

                default:
                    throw new ArgumentException("Invalid SeekOrgin", "origin");
            }

            position = newPosition;
            return position;
        }

        public override void SetLength(long value)
        {
            throw new System.NotImplementedException("The method is not implemented.");
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            if (disposed)
            {
                throw new ObjectDisposedException(typeof(NonContiguousMemoryStream).ToString());
            }

            if (buffer == null)
            {
                throw new ArgumentNullException("buffer");
            }

            if (offset < 0 || offset >= buffer.Length)
            {
                throw new ArgumentOutOfRangeException("offset", offset, "0 <= offset < buffer.Length");
            }

            if (count < 0 || count > buffer.Length)
            {
                throw new ArgumentOutOfRangeException("count", count, "0 <= count < buffer.Length");
            }

            if ((offset + count) > buffer.Length)
            {
                throw new ArgumentOutOfRangeException("offset + count", (offset + count), "(offset + count) <= buffer.Length");
            }

            while ((position + count) > (buffersStore.Count * bufferStoreSize))
            {
                buffersStore.Add(new byte[bufferStoreSize]);
            }

            int end = offset + count;
            int buffersStoreIndex = (int)(position / bufferStoreSize);
            int bufferStoreOffset = (int)(position % bufferStoreSize);
            byte[] bufferStore = buffersStore[buffersStoreIndex];

            while (offset < end)
            {
                int run = Math.Min(end - offset, bufferStoreSize - bufferStoreOffset);
                Buffer.BlockCopy(buffer, offset, bufferStore, bufferStoreOffset, run);
                offset += run;
                bufferStoreOffset += run;

                if (bufferStoreOffset == bufferStoreSize && offset < end)
                {
                    buffersStoreIndex += 1;
                    bufferStore = buffersStore[buffersStoreIndex];
                    bufferStoreOffset = 0;
                }
            }

            position += count;

            if (position > length)
            {
                length = position;
            }
        }

        protected override void Dispose(bool disposing)
        {
            if (disposed)
            {
                return;
            }

            try
            {
                if (disposing)
                {
                    disposed = true;
                    length = 0;
                    position = 0;
                    buffersStore.Clear();
                    buffersStore = null;
                }
            }

            finally
            {
                base.Dispose(disposing);
            }
        }
    }
}
